/* ========================================================================
 * Copyright 2012 Chris Moos
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Converted to Arduino library by Craig A. Lindley
 * Version: 0.5
 * Last Update: 01/08/2013
 * ========================================================================
 */

#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>

#include "AVR_OS.h"

// Set to the timer you want to use
#define CONFIG_AVR_TIMER 1

#define MAX_TASKS 5
#define TICK_INTERVAL 10
#define TASK_STACK_SIZE 256
#define TASK_FLAG_RUNNING 1
#define TASK_FLAG_DONE 2

#define BV(bit) (1 << bit)
#define BIT_CLEAR(n, bit) n &= ~BV(bit);
#define BIT_SET(n, bit) n |= BV(bit);
#define BIT_ISSET(n, bit) (n & BV(bit))

struct task_definition {
    void *arg;
    void *address;
    uint8_t flags;
    void *original_sp;
    void *saved_sp;
    uint32_t start_delay_secs;
    uint32_t delayMillis;
};
typedef struct task_definition task_definition;

#define xstr(s) str(s)
#define str(s) #s

#define PRINT_STACK_POINTER \
    asm volatile("push r0\nin r0, __SP_L__\nsts stackPointer, r0\nin r0, __SP_H__\nsts stackPointer+1, r0\npop r0"); printf("stack pointer: %04x\n", stackPointer);

#define SAVE_CONTEXT(sp) \
    asm volatile( \
        "push r0\n" \
        "in r0, __SREG__\n" \
        "cli\n" \
        "push r0\n" \
        "push r1\n" \
        "push r2\n" \
        "push r3\n" \
        "push r4\n" \
        "push r5\n" \
        "push r6\n" \
        "push r7\n" \
        "push r8\n" \
        "push r9\n" \
        "push r10\n" \
        "push r11\n" \
        "push r12\n" \
        "push r13\n" \
        "push r14\n" \
        "push r15\n" \
        "push r16\n" \
        "push r17\n" \
        "push r18\n" \
        "push r19\n" \
        "push r20\n" \
        "push r21\n" \
        "push r22\n" \
        "push r23\n" \
        "push r24\n" \
        "push r25\n" \
        "push r26\n" \
        "push r27\n" \
        "push r28\n" \
        "push r29\n" \
        "push r30\n" \
        "push r31\n" \
        "in r0, __SP_L__\n" \
        "sts "xstr(sp)", r0\n" \
        "in r0, __SP_H__\n" \
        "sts "xstr(sp+1)", r0\n" \
    );

#define RESTORE_CONTEXT(sp) \
    asm volatile( \
        "lds r30, "xstr(sp)"\n" \
        "lds r31, "xstr(sp+1)"\n" \
        "out __SP_L__, r30\n" \
        "out __SP_H__, r31\n" \
        "pop r31\n" \
        "pop r30\n" \
        "pop r29\n" \
        "pop r28\n" \
        "pop r27\n" \
        "pop r26\n" \
        "pop r25\n" \
        "pop r24\n" \
        "pop r23\n" \
        "pop r22\n" \
        "pop r21\n" \
        "pop r20\n" \
        "pop r19\n" \
        "pop r18\n" \
        "pop r17\n" \
        "pop r16\n" \
        "pop r15\n" \
        "pop r14\n" \
        "pop r13\n" \
        "pop r12\n" \
        "pop r11\n" \
        "pop r10\n" \
        "pop r9\n" \
        "pop r8\n" \
        "pop r7\n" \
        "pop r6\n" \
        "pop r5\n" \
        "pop r4\n" \
        "pop r3\n" \
        "pop r2\n" \
        "pop r1\n" \
        "pop r0\n" \
        "out __SREG__, r0\n" \
        "pop r0\n" \
    );


task_definition tasks[MAX_TASKS];

volatile int cur_task = -1;
volatile int num_tasks = 0;

uint32_t uptime_secs = 0;
uint64_t uptime_millis = 0;

void *tempSp;
uint8_t *stackTop, *tasksStackTop;

// Forward References
void _os_platform_do_something_else();
void _os_task_delay_starter(void *arg);
int _os_platform_schedule_task(os_task_function function, void *arg, uint16_t start_delay_secs, uint8_t clear_interrupts);
void do_something_else();
void _os_platform_spinlock_acquire(spinlock_t *lock);
void _os_platform_spinlock_release(spinlock_t *lock);
void mutex_acquire(mutex_t *mutex);
void mutex_release(mutex_t *mutex);
void _os_platform_mutex_acquire(mutex_t *mutex);
void _os_platform_mutex_release(mutex_t *mutex);
int start_task(os_task_function addr, void *arg, uint16_t start_delay);
void _os_sleep(uint16_t millis);
void _os_platform_loop();


AVR_OS::AVR_OS() {
	
	cli();

#if CONFIG_AVR_TIMER == 1

	// clear on compare match
	//TCCR1A = _BV(COM1A1);

	// clkio/256, CTC, TOP = OCR1A
	TCCR1B = _BV(CS12) | _BV(WGM12);

	OCR1A = 625;// F_CPU / 256 / (1000 / TICK_INTERVAL);

	TIMSK1 = _BV(OCIE1A);

#elif CONFIG_AVR_TIMER == 0
    
	TCCR0A = _BV(COM0A1) | _BV(WGM01);
	TCCR0B = _BV(CS02) | _BV(CS00);
	OCR0A = F_CPU / 1024 / (1000 / TICK_INTERVAL);
	TIMSK0 = _BV(OCIE0A);

#elif CONFIG_AVR_TIMER == 2
    
	TCCR2A = _BV(COM2A1) | _BV(WGM11);
	TCCR2B = _BV(CS22) | _BV(CS21) | _BV(CS20);
	OCR2A = F_CPU / 1024 / (1000 / TICK_INTERVAL);
	TIMSK2 = _BV(OCIE2A);

#endif

	_os_platform_schedule_task(_os_task_delay_starter, NULL, 0, 0);
}

void AVR_OS::os_loop() {
    
	_os_platform_loop();
}

int AVR_OS::os_schedule_task(os_task_function function, void *arg, uint16_t start_delay_secs) {

	return _os_platform_schedule_task(function, arg, start_delay_secs, 1);
}

void AVR_OS::os_exit_task() {
    
	BIT_CLEAR(tasks[cur_task].flags, TASK_FLAG_RUNNING);
	BIT_SET(tasks[cur_task].flags, TASK_FLAG_DONE);
	cur_task = -1;

	do_something_else();
}



void AVR_OS::os_yield() {

	_os_platform_do_something_else();
}


void AVR_OS::os_sleep(uint16_t millis) {

	_os_sleep(millis);
}

uint32_t AVR_OS::os_get_uptime() {
    
	return uptime_secs;
}

uint64_t AVR_OS::os_get_uptime_millis() {
    
	return uptime_millis;
}

void AVR_OS::spinlock_init(spinlock_t *lock) {

	*lock = 0;
}

void AVR_OS::spinlock_acquire(spinlock_t *lock) {
    
	_os_platform_spinlock_acquire(lock);
}

void AVR_OS::spinlock_release(spinlock_t *lock) {
    
	_os_platform_spinlock_release(lock); 
}


void AVR_OS::mutex_init(mutex_t *mutex) {
    
	memset((void*)mutex, 0, sizeof(mutex_t));
}

void AVR_OS::mutex_acquire(mutex_t *mutex) {
    
	_os_platform_mutex_acquire(mutex);
}

void AVR_OS::mutex_release(mutex_t *mutex) {
    
	_os_platform_mutex_release(mutex);
}



void _os_platform_sleep() {
    sleep_cpu();
}

int _os_platform_schedule_task(os_task_function function, void *arg, uint16_t start_delay_secs, uint8_t clear_interrupts) {

    if(clear_interrupts == 1 && (SREG & (1 << 7)) == (1 << 7)) {
        cli();
        int ret = start_task(function, arg, start_delay_secs);
        
        sei();
        return ret;
    }
    else {
        return start_task(function, arg, start_delay_secs);
    }
}

void _os_platform_switch_tasks() {
    
    if(num_tasks == 0) return;


    uint8_t searched = num_tasks;
    while(1) {
        cur_task++;
        if(cur_task == num_tasks) cur_task = 0;
        if(searched == 0) {
            cur_task = -1;
            sei();
            while(1)
                sleep_cpu();
        }
        searched--;

        if(BIT_ISSET(tasks[cur_task].flags, TASK_FLAG_RUNNING)) {
            if(tasks[cur_task].delayMillis <= 0) {
                tempSp = tasks[cur_task].saved_sp;
                break;
            }
        }
        else if(BIT_ISSET(tasks[cur_task].flags, TASK_FLAG_DONE)) {
            continue;
        }
        else if(tasks[cur_task].start_delay_secs == 0) {
            uint8_t *top;

            if(tasks[cur_task].original_sp != NULL) {
                top = (uint8_t *) tasks[cur_task].original_sp; 
            }
            else {
                top = (uint8_t *) tasksStackTop;
                tasksStackTop -= TASK_STACK_SIZE;
                tasks[cur_task].original_sp = top;
            }
            memset(top - TASK_STACK_SIZE, 0, TASK_STACK_SIZE);

            *(top) = ((uint16_t)tasks[cur_task].address);
            *(top - 1) = (((uint16_t)tasks[cur_task].address) >> 8);

            uint8_t *registerStartAddr = top - 3;
            *(registerStartAddr - 24) = ((uint16_t)tasks[cur_task].arg);
            *(registerStartAddr - 25) = (((uint16_t)tasks[cur_task].arg) >> 8);

            tasks[cur_task].saved_sp = top - 35;
            BIT_SET(tasks[cur_task].flags, TASK_FLAG_RUNNING);
            tempSp = tasks[cur_task].saved_sp;

            break;
        }
    }
    
}

void _os_platform_do_something_else() {
    cli();

    SAVE_CONTEXT(tempSp)

    // set the stack top for the ISR
    asm volatile(
        "lds r16, stackTop\n"
        "lds r17, stackTop+1\n"
        "out __SP_L__, r16\n"
        "out __SP_H__, r17\n"
    );

    if(cur_task != -1) {
        tasks[cur_task].saved_sp = tempSp;
    }

    _os_platform_switch_tasks();

    RESTORE_CONTEXT(tempSp)
    sei();
}

void _os_platform_update_delay_millis() {
    int x;
    for(x = 0; x < num_tasks; x++) {
        if(BIT_ISSET(tasks[x].flags, TASK_FLAG_RUNNING) && tasks[x].delayMillis > 0) {
            tasks[x].delayMillis -= TICK_INTERVAL;
        }
    }
}

#if CONFIG_AVR_TIMER == 0
ISR(TIMER0_COMPA_vect, ISR_NAKED) {
#elif CONFIG_AVR_TIMER == 1
ISR(TIMER1_COMPA_vect, ISR_NAKED) {
#elif CONFIG_AVR_TIMER == 2
ISR(TIMER2_COMPA_vect, ISR_NAKED) {
#endif
    SAVE_CONTEXT(tempSp)

    // set the stack top for the ISR
    asm volatile(
        "lds r16, stackTop\n"
        "lds r17, stackTop+1\n"
        "out __SP_L__, r16\n"
        "out __SP_H__, r17\n"
    );

    uptime_millis += TICK_INTERVAL;

    if(cur_task != -1) {
        tasks[cur_task].saved_sp = tempSp;
    }


    _os_platform_update_delay_millis();
    _os_platform_switch_tasks();

    RESTORE_CONTEXT(tempSp)
    asm volatile("reti");
}

void _os_platform_spinlock_acquire(spinlock_t *lock) {
    while(1) {
        if(*lock == 0) {
            cli();
            if(*lock == 0) {
                *lock = 1;
                sei();
                return;
            }
            sei();
        }
    }
}

void _os_platform_spinlock_release(spinlock_t *lock) {
    cli();
    *lock = 0;
    sei();
}

void _os_platform_mutex_acquire(mutex_t *mutex) {
    uint8_t pending = 0;

    while(1) {
        cli();
        if(mutex->value == 0 && (mutex->wait == 0 || mutex->wait == cur_task)) {
            mutex->value = 1; 
            mutex->wait = 0;
            sei();
            return;
        }
        else {
            if(pending == 0 && mutex->wait == 0) {
                pending = 1; 
                mutex->wait = cur_task;
            }
            _os_platform_do_something_else();
        }
    }
}

void _os_platform_mutex_release(mutex_t *mutex) {
    mutex->value = 0;
}


void _os_task_delay_starter(void *arg) {
    int x;
    while(1) {
        _os_sleep(1000);
        uptime_secs++;

        for(x = 0; x < num_tasks; x++) {
            if(!BIT_ISSET(tasks[x].flags, TASK_FLAG_RUNNING) && !BIT_ISSET(tasks[x].flags, TASK_FLAG_DONE) &&
                tasks[x].start_delay_secs > 0) {

                tasks[x].start_delay_secs -= 1;

            }
        }
    }
}

void do_something_else() {
    _os_platform_do_something_else();
}

void mutex_acquire(mutex_t *mutex) {
    _os_platform_mutex_acquire(mutex);
}

int start_task(os_task_function addr, void *arg, uint16_t start_delay) {
    if(num_tasks == MAX_TASKS) {
        // search for a task that is marked as done, take it over
        int x;
        for(x = 0; x < num_tasks; x++) {
            if(BIT_ISSET(tasks[x].flags, TASK_FLAG_DONE)) {
                tasks[x].address = (void *) addr;
                tasks[x].arg = arg;
                tasks[x].start_delay_secs = start_delay;

                // reset flags
                // running = 0
                // done = 0
                tasks[x].flags = 0;
                tasks[x].saved_sp = NULL;
                tasks[x].delayMillis = 0;
                return 0;
            }
        }
        return -1;
    }
    else {
        tasks[num_tasks].address = (void *) addr;
        tasks[num_tasks].arg = arg;
        tasks[num_tasks].start_delay_secs = start_delay;
        num_tasks++;
        return 0;
    }
}

void _os_sleep(uint16_t millis) {
    tasks[cur_task].delayMillis = millis;
    do_something_else();
}

void _os_platform_loop() {
    asm volatile("in r0, __SP_L__\nsts stackTop, r0\nin r0, __SP_H__\nsts stackTop+1, r0");

    // reserve for the while loop
    stackTop -= 32;
    tasksStackTop = stackTop - 256;

    _os_platform_do_something_else();
}


